@@ -3,6 +3,8 @@
 
 #include "mobipocket.h"
 #include "decompressor.h"
+#include "pdb_p.h"
+#include "qmobipocket_debug.h"
 
 #include <QBuffer>
 #include <QIODevice>
@@ -18,75 +20,6 @@
 namespace Mobipocket
 {
 
-struct PDBPrivate {
-    QList<quint32> recordOffsets;
-    QIODevice *device;
-    QString fileType;
-    quint16 nrecords;
-    bool valid;
-
-    void init();
-};
-
-void PDBPrivate::init()
-{
-    valid = true;
-    quint16 word;
-    quint32 dword;
-    if (!device->seek(0x3c))
-        goto fail;
-    fileType = QString::fromLatin1(device->read(8));
-
-    if (!device->seek(0x4c))
-        goto fail;
-    device->read((char *)&word, 2);
-    nrecords = qFromBigEndian(word);
-
-    for (int i = 0; i < nrecords; i++) {
-        device->read((char *)&dword, 4);
-        recordOffsets.append(qFromBigEndian(dword));
-        device->read((char *)&dword, 4);
-    }
-    return;
-fail:
-    valid = false;
-}
-
-PDB::PDB(QIODevice *device)
-    : d(new PDBPrivate)
-{
-    d->device = device;
-    d->init();
-}
-
-PDB::~PDB()
-{
-    delete d;
-}
-
-QByteArray PDB::getRecord(int i) const
-{
-    if (i >= d->nrecords)
-        return QByteArray();
-    quint32 offset = d->recordOffsets[i];
-    bool last = (i == (d->nrecords - 1));
-    if (!d->device->seek(offset))
-        return QByteArray();
-    if (last)
-        return d->device->readAll();
-    return d->device->read(d->recordOffsets[i + 1] - offset);
-}
-
-bool PDB::isValid() const
-{
-    return d->valid;
-}
-
-int PDB::recordCount() const
-{
-    return d->nrecords;
-}
-
 struct DocumentPrivate 
 {
     DocumentPrivate(QIODevice *d)
@@ -119,7 +52,6 @@ struct DocumentPrivate
     void parseEXTH(const QByteArray &data);
     void parseHtmlHead(const QString &data);
     QString readStringRecord(const QByteArray &data);
-    QImage getImageFromRecord(int recnum);
 };
 
 void DocumentPrivate::parseHtmlHead(const QString &data)
@@ -161,6 +93,11 @@ namespace {
         quint32 huff_ofs = qFromBigEndian<quint32>(header.constData() + 0x70);
         quint32 huff_num = qFromBigEndian<quint32>(header.constData() + 0x74);
 
+        // Check for overflow and out-of-bounds access
+        if (((huff_ofs + huff_num) < huff_num) || ((huff_ofs + huff_num) > pdb.recordCount())) {
+            return {};
+        }
+
         QVector<QByteArray> records(huff_num);
         for (quint32 i = 0; i < huff_num; i++) {
             if (auto r = pdb.getRecord(huff_ofs + i); r.isNull()) {
@@ -197,10 +134,9 @@ void DocumentPrivate::init()
     if (encoding == 65001) {
         toUtf16 = QStringDecoder(QStringDecoder::Utf8);
     } else {
-        const auto converterEncoding = QStringConverter::encodingForName("cp1252");
-        if (converterEncoding) {
-            toUtf16 = QStringDecoder(*converterEncoding);
-        } else {
+        toUtf16 = QStringDecoder("windows-1252");
+        if (!toUtf16.isValid()) {
+            qCWarning(QMOBIPOCKET_LOG) << "Text codec \"windows-1252\" not supported by Qt library, falling back to Latin1";
             toUtf16 = QStringDecoder(QStringConverter::Latin1);
         }
     }
@@ -213,9 +149,11 @@ void DocumentPrivate::init()
     if (mhead.size() >= 92)
         parseEXTH(mhead);
 
-    quint32 exthoffs = qFromBigEndian<quint32>(mhead.constData() + 20);
-    if ((mhead.size() >= 244) && ((exthoffs + 16) > 244)) {
-        extraflags = qFromBigEndian<quint32>(mhead.constData() + 240);
+    if (mhead.size() >= 244) {
+        quint32 exthoffs = qFromBigEndian<quint32>(mhead.constData() + 20);
+        if ((exthoffs + 16) > 244) {
+            extraflags = qFromBigEndian<quint32>(mhead.constData() + 240);
+        }
     }
 
     // try getting metadata from HTML if nothing or only title was recovered from MOBI and EXTH records
@@ -253,12 +191,6 @@ QString DocumentPrivate::readStringRecord(const QByteArray &data)
 #endif
 }
 
-QImage DocumentPrivate::getImageFromRecord(int i)
-{
-    QByteArray rec = pdb.getRecord(i);
-    return (rec.isNull()) ? QImage() : QImage::fromData(rec);
-}
-
 void DocumentPrivate::parseEXTH(const QByteArray &data)
 {
     // try to get name
@@ -275,6 +207,9 @@ void DocumentPrivate::parseEXTH(const QByteArray &data)
     }
 
     quint32 exthoffs = qFromBigEndian<quint32>(data.constData() + 20);
+    if (exthoffs + 28 > quint32(data.size())) {
+        return;
+    }
 
     if (data.mid(exthoffs + 16, 4) != "EXTH")
         return;
@@ -395,7 +330,7 @@ QString Document::text(int size) const
 #if QT_VERSION > QT_VERSION_CHECK(6, 0, 0)
     return d->toUtf16(whole);
 #else
-    return d->codec->toUnicode(whole);
+    return d->codec ? d->codec->toUnicode(whole) : QString();
 #endif
 }
 
@@ -414,7 +349,14 @@ QImage Document::getImage(int i) const
 {
     if (!d->firstImageRecord)
         d->findFirstImage();
-    return d->getImageFromRecord(d->firstImageRecord + i);
+
+    if ((i < 0) || (i > std::numeric_limits<quint16>::max()) //
+        || (d->firstImageRecord + i) >= d->pdb.recordCount()) {
+        return {};
+    }
+
+    QByteArray rec = d->pdb.getRecord(d->firstImageRecord + i);
+    return (rec.isNull()) ? QImage() : QImage::fromData(rec);
 }
 
 QMap<Document::MetaKey, QString> Document::metadata() const
@@ -429,13 +371,12 @@ bool Document::hasDRM() const
 
 QImage Document::thumbnail() const
 {
-    if (!d->firstImageRecord)
-        d->findFirstImage();
-    QImage img = d->getImageFromRecord(d->thumbnailIndex + d->firstImageRecord);
+    QImage img = getImage(d->thumbnailIndex);
+
     // does not work, try first image
     if (img.isNull() && d->thumbnailIndex) {
         d->thumbnailIndex = 0;
-        img = d->getImageFromRecord(d->firstImageRecord);
+        img = getImage(0);
     }
     return img;
 }
